use std::{io::Write, process::Child, sync::RwLock};

use crate::{backup::MetaManager, server::Metadata};

pub struct ServerProcess {
    child: RwLock<Child>,
    pub backups: RwLock<MetaManager<Metadata>>,
}

impl ServerProcess {
    pub fn new(manager: MetaManager<Metadata>, child: Child) -> ServerProcess {
        ServerProcess {
            child: RwLock::new(child),
            backups: RwLock::new(manager),
        }
    }

    pub fn send_command(&self, cmd: &str) -> std::io::Result<()> {
        match cmd.trim() {
            "stop" | "exit" => self.stop()?,
            "backup" => self.backup()?,
            s => self.custom(s)?,
        }

        Ok(())
    }

    fn custom(&self, cmd: &str) -> std::io::Result<()> {
        let child = self.child.write().unwrap();
        let mut stdin = child.stdin.as_ref().unwrap();
        stdin.write_all(format!("{}\n", cmd.trim()).as_bytes())?;
        stdin.flush()?;

        Ok(())
    }

    pub fn stop(&self) -> std::io::Result<()> {
        self.custom("stop")?;

        self.child.write().unwrap().wait()?;

        Ok(())
    }

    pub fn kill(&self) -> std::io::Result<()> {
        self.child.write().unwrap().kill()
    }

    /// Perform a backup by disabling the server's save feature and flushing its data, before
    /// creating an archive file.
    pub fn backup(&self) -> std::io::Result<()> {
        // We explicitely lock this entire function to prevent parallel backups
        let mut backups = self.backups.write().unwrap();

        let layer_name = String::from(backups.next_scheduled_layer().unwrap());
        self.custom(&format!("say starting backup for layer '{}'", layer_name))?;

        // Make sure the server isn't modifying the files during the backup
        self.custom("save-off")?;
        self.custom("save-all")?;

        // TODO implement a better mechanism
        // We wait some time to (hopefully) ensure the save-all call has completed
        std::thread::sleep(std::time::Duration::from_secs(10));

        let start_time = chrono::offset::Utc::now();
        let res = backups.perform_backup_cycle();

        // The server's save feature needs to be enabled again even if the archive failed to create
        self.custom("save-on")?;
        self.custom("save-all")?;

        let duration = chrono::offset::Utc::now() - start_time;
        let duration_str = format!(
            "{}m{}s",
            duration.num_seconds() / 60,
            duration.num_seconds() % 60
        );

        if res.is_ok() {
            self.custom(&format!(
                "say backup created for layer '{}' in {}",
                layer_name, duration_str
            ))?;
        } else {
            self.custom(&format!(
                "an error occured after {} while creating backup for layer '{}'",
                duration_str, layer_name
            ))?;
        }

        res
    }
}
